@@ -324,7 +324,7 @@ class VariableSubclassTestCases(object):
                        expected[...],
                        expected.squeeze(),
                        expected.isel(x=slice(None)),
-                       expected.expand_dims({'x': 3}),
+                       expected.set_dims({'x': 3}),
                        expected.copy(deep=True),
                        expected.copy(deep=False)]:
 
@@ -609,11 +609,14 @@ class TestVariable(TestCase, VariableSubclassTestCases):
     def test_as_variable(self):
         data = np.arange(10)
         expected = Variable('x', data)
+        expected_extra = Variable('x', data, attrs={'myattr': 'val'},
+                                  encoding={'scale_factor': 1})
 
         self.assertVariableIdentical(expected, as_variable(expected))
 
         ds = Dataset({'x': expected})
-        self.assertVariableIdentical(expected, as_variable(ds['x']).to_base_variable())
+        var = as_variable(ds['x']).to_base_variable()
+        self.assertVariableIdentical(expected, var)
         self.assertNotIsInstance(ds['x'], Variable)
         self.assertIsInstance(as_variable(ds['x']), Variable)
 
@@ -621,8 +624,20 @@ class TestVariable(TestCase, VariableSubclassTestCases):
         fake_xarray = FakeVariable(expected.values, expected.dims)
         self.assertVariableIdentical(expected, as_variable(fake_xarray))
 
-        xarray_tuple = (expected.dims, expected.values)
-        self.assertVariableIdentical(expected, as_variable(xarray_tuple))
+        FakeVariable = namedtuple('FakeVariable', 'data dims')
+        fake_xarray = FakeVariable(expected.data, expected.dims)
+        self.assertVariableIdentical(expected, as_variable(fake_xarray))
+
+        FakeVariable = namedtuple('FakeVariable',
+                                  'data values dims attrs encoding')
+        fake_xarray = FakeVariable(expected_extra.data, expected_extra.values,
+                                   expected_extra.dims, expected_extra.attrs,
+                                   expected_extra.encoding)
+        self.assertVariableIdentical(expected_extra, as_variable(fake_xarray))
+
+        xarray_tuple = (expected_extra.dims, expected_extra.values,
+                        expected_extra.attrs, expected_extra.encoding)
+        self.assertVariableIdentical(expected_extra, as_variable(xarray_tuple))
 
         with self.assertRaisesRegexp(TypeError, 'tuples to convert'):
             as_variable(tuple(data))
@@ -637,6 +652,15 @@ class TestVariable(TestCase, VariableSubclassTestCases):
         expected = Variable([], 0)
         self.assertVariableIdentical(expected, actual)
 
+        data = np.arange(9).reshape((3, 3))
+        expected = Variable(('x', 'y'), data)
+        with self.assertRaisesRegexp(
+                ValueError, 'without explicit dimension names'):
+            as_variable(data, name='x')
+        with self.assertRaisesRegexp(
+                ValueError, 'has more than 1-dimension'):
+            as_variable(expected, name='x')
+
     def test_repr(self):
         v = Variable(['time', 'x'], [[1, 2, 3], [4, 5, 6]], {'foo': 'bar'})
         expected = dedent("""
@@ -644,7 +668,7 @@ class TestVariable(TestCase, VariableSubclassTestCases):
         array([[1, 2, 3],
                [4, 5, 6]])
         Attributes:
-            foo: bar
+            foo:      bar
         """).strip()
         self.assertEqual(expected, repr(v))
 
@@ -818,30 +842,30 @@ class TestVariable(TestCase, VariableSubclassTestCases):
         with self.assertRaisesRegexp(ValueError, 'not found in array dim'):
             v.get_axis_num('foobar')
 
-    def test_expand_dims(self):
+    def test_set_dims(self):
         v = Variable(['x'], [0, 1])
-        actual = v.expand_dims(['x', 'y'])
+        actual = v.set_dims(['x', 'y'])
         expected = Variable(['x', 'y'], [[0], [1]])
         self.assertVariableIdentical(actual, expected)
 
-        actual = v.expand_dims(['y', 'x'])
+        actual = v.set_dims(['y', 'x'])
         self.assertVariableIdentical(actual, expected.T)
 
-        actual = v.expand_dims(OrderedDict([('x', 2), ('y', 2)]))
+        actual = v.set_dims(OrderedDict([('x', 2), ('y', 2)]))
         expected = Variable(['x', 'y'], [[0, 0], [1, 1]])
         self.assertVariableIdentical(actual, expected)
 
         v = Variable(['foo'], [0, 1])
-        actual = v.expand_dims('foo')
+        actual = v.set_dims('foo')
         expected = v
         self.assertVariableIdentical(actual, expected)
 
         with self.assertRaisesRegexp(ValueError, 'must be a superset'):
-            v.expand_dims(['z'])
+            v.set_dims(['z'])
 
-    def test_expand_dims_object_dtype(self):
+    def test_set_dims_object_dtype(self):
         v = Variable([], ('a', 1))
-        actual = v.expand_dims(('x',), (3,))
+        actual = v.set_dims(('x',), (3,))
         exp_values = np.empty((3,), dtype=object)
         for i in range(3):
             exp_values[i] = ('a', 1)
