@@ -13,6 +13,10 @@ priority order:
 Once NumPy 1.10 comes out with support for overriding ufuncs, this module will
 hopefully no longer be necessary.
 """
+from __future__ import absolute_import
+from __future__ import division
+from __future__ import print_function
+
 import numpy as _np
 
 from .core.variable import Variable as _Variable
@@ -21,7 +25,7 @@ from .core.dataarray import DataArray as _DataArray
 from .core.groupby import GroupBy as _GroupBy
 
 from .core.pycompat import dask_array_type as _dask_array_type
-from .core.ops import _dask_or_eager_func
+from .core.duck_array_ops import _dask_or_eager_func
 
 
 _xarray_types = (_Variable, _DataArray, _Dataset, _GroupBy)
